import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.media.jai.Histogram;
import java.awt.Image;
import java.awt.image.*;

/** Class that contains all the algorithm functionality for video shot detection.
 * 
 * @version 1.0 01.06.06
 * @author Vegard Kosmo
 * @see     Videosegmenter
 * @see     GlobalHistogram
 * @see     LocalHistogram
 * @see     EdgeDetector
 * @see     EdgeHistogram
 * @see     ResizedImageObject
 *
 */
public class Algorithm{
	// Main application reference
	Videosegmenter main;
		
	// GlobalHistogram reference
	GlobalHistogram globalHistogram;
	
	// LocalHistogram reference
	LocalHistogram localHistogram;
	
	// Edge detector reference
	EdgeDetector edgeDetector;	
	
	// EdgeHistogram reference
	EdgeHistogram edgeHistogram;
	
	// ResizedImageObject reference
	ResizedImageObject resizedImageObject1;
	ResizedImageObject resizedImageObject2;
	
	// Algorithm variables
	private int framePick;
	private PlanarImage planarImage;
	private PlanarImage edgeImage;
	private Image image;
	private Histogram histogram;
	private Histogram[] localHistogramList;
	private Histogram[] edgeHistogramList;
	private Histogram[] histogramArray;
	private Histogram[][] localHistogramArray;
	private Histogram[][] edgeHistogramArray;
	private Image[] imageArray;	
	public double globalColorThreshold;
	public double localColorThreshold;
	public double edgeColorThreshold;
	private double percent;
	private int imageWidth;
	private int imageHeight;
	private int numberOfPixels;
	private int imageCounter;

	
	/** Standard constructor. Creates histograms and structures for storing the results.
	 * 
	 * @param main Main class reference
	 */
	public Algorithm(Videosegmenter main){
		this.main = main;
		percent = 0.95;
		globalHistogram = new GlobalHistogram();
		localHistogram = new LocalHistogram();
		edgeHistogram = new EdgeHistogram();
		main.buildGlobalHistogramComparator();
		main.buildLocalHistogramComparator();
		main.buildEdegHistogramComparator();
		histogramArray = new Histogram[2];
		localHistogramArray = new Histogram[2][9];
		edgeHistogramArray = new Histogram[2][9];
		imageArray = new Image[2];
		edgeDetector = new EdgeDetector();
		imageCounter = 0;		
	}
		
		
	/** Method to run one of the algorithms, boolean values decide which one to be run
	 * 
	 */
	public void runAlgorithm(){
		
		// The global color histogram algorithm	
		if(main.algorithm1){
						
			for(int i = 0; i <= ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond())); i++){
				
				// Picks out the first frame, the last frame and a given number of frames per second									
				// Picks the first frame
				if(i == 0){
					framePick = 1;
					image = main.frameSelector.grabFrame(framePick);
					imageArray[0] = image;
					planarImage = JAI.create("AWTImage", image);
					histogram = globalHistogram.createGlobalHistogram(planarImage);
					histogramArray[0] = histogram;
					
					// Sets the image stats
					imageWidth = planarImage.getWidth();
					imageHeight = planarImage.getHeight();
					numberOfPixels = imageWidth*imageHeight;
					globalColorThreshold = numberOfPixels*percent;
															
					// Adds the first image to the GUI
					BufferedImage bufferedImage1 = main.imageResizer.resizeImage((Image)imageArray[0], imageWidth, imageHeight);
					resizedImageObject1 = new ResizedImageObject(bufferedImage1, 1, 1);
					main.gui.addImage(resizedImageObject1);					
				}
				
				// Picks the last frame
				else if(i == ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond()))){
					framePick = ((i*(main.mediaPlayer.getFrequency())));
										
					// Creates histogram 
					image = main.frameSelector.grabFrame(framePick);
					imageArray[1] = image;
					planarImage = JAI.create("AWTImage", image);
					histogram = globalHistogram.createGlobalHistogram(planarImage);
					histogramArray[1] = histogram;
				}
				
				// Picks frames between the first frame and the last frame
				else{
					framePick = ((i*(main.mediaPlayer.getFrequency())));
					
					// Creates histogram 
					image = main.frameSelector.grabFrame(framePick);
					imageArray[1] = image;
					planarImage = JAI.create("AWTImage", image);
					histogram = globalHistogram.createGlobalHistogram(planarImage);
					histogramArray[1] = histogram;					
				}
				
				// Checks if the histogram difference is greater than the threshold value for a cut				
				if(i > 0){
					
					if(main.globalHistogramComparator.compareGlobalHistograms((Histogram)histogramArray[0], (Histogram)histogramArray[1])){
						BufferedImage bufferedImage1 = main.imageResizer.resizeImage((Image)imageArray[0], imageWidth, imageHeight);
						BufferedImage bufferedImage2 = main.imageResizer.resizeImage((Image)imageArray[1], imageWidth, imageHeight);
						resizedImageObject1 = new ResizedImageObject(bufferedImage1, i-1, (i-1)*(main.mediaPlayer.getFrequency()));
						resizedImageObject2 = new ResizedImageObject(bufferedImage2, i, (i)*(main.mediaPlayer.getFrequency()));
												
						// Adds the remaining cut images
						main.gui.addImage(resizedImageObject2);
						imageCounter++;						
					}
					
					// Sets the shot window properties after the last iteration
					if(i == ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond()))){
												
						// Sets the number of rows in the GUI
						double doubleImageCounter = (double)imageCounter;						
						double numberOfRows = (doubleImageCounter+1)/7;						
						double rowsCeil = Math.ceil(numberOfRows);						
						int integerNumberOfRows = (int)rowsCeil;						
						
						int imagePanelWidth;
						
						if(imageCounter >= 6){
							imagePanelWidth = (((resizedImageObject1.getBufferedImage().getWidth())*7)+38);							
						}
						
						else{
							imagePanelWidth = (((resizedImageObject1.getBufferedImage().getWidth())*(imageCounter+1))+38);							
						}
						
						int imagePanelHeight = imagePanelHeight = (((resizedImageObject1.getBufferedImage().getHeight())*integerNumberOfRows));
						
						main.gui.gridLayout.setRows(integerNumberOfRows);
						main.gui.imagePanel.setSize(imagePanelWidth, imagePanelHeight);
						
						// Sets the scrolling in the GUI if necessary
						if(main.gui.imagePanel.getHeight() >= 575){
							main.shotWindow.setSize(main.gui.imagePanel.getWidth(), 575);
						}
						else{
							main.shotWindow.setSize(main.gui.imagePanel.getWidth(), main.gui.imagePanel.getHeight()+30);							
						}
						
						// Makes the shot window visible						
						main.gui.shotWindow.setTitle("Video shot viewer - This video contains "+(imageCounter+1)+" shots");
						main.shotWindow.setVisible(true);
					}
					
					// Changes objects for the next iteration
					imageArray[0] = (Image)imageArray[1];
					histogramArray[0] = (Histogram)histogramArray[1];					
				}																
			}						
		}
		
		// The local color histogram algorithm
		else if(main.algorithm2){
			
			for(int i = 0; i <= ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond())); i++){
								
				// Picks out the first frame, the last frame and a given number of frames per second									
				// Picks the first frame
				if(i == 0){
					framePick = 1;
					image = main.frameSelector.grabFrame(framePick);
					imageArray[0] = image;
					planarImage = JAI.create("AWTImage", image);
					localHistogramList = localHistogram.createLocalHistogram(planarImage);
					
					for(int m = 0; m < localHistogramList.length; m++){
						localHistogramArray[0][m] = (Histogram)localHistogramList[m];						
					}
										
					// Sets the image stats
					imageWidth = planarImage.getWidth();
					imageHeight = planarImage.getHeight();
					numberOfPixels = imageWidth*imageHeight;
					localColorThreshold = (numberOfPixels/9)*percent;
					
					// Adds the first image to the GUI
					BufferedImage bufferedImage1 = main.imageResizer.resizeImage((Image)imageArray[0], imageWidth, imageHeight);
					resizedImageObject1 = new ResizedImageObject(bufferedImage1, 1, 1);
					main.gui.addImage(resizedImageObject1);					
				}
				
				// Picks the last frame
				else if(i == ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond()))){
					framePick = (i*(main.mediaPlayer.getFrequency()));
										
					// Creates histogram 
					image = main.frameSelector.grabFrame(framePick);
					imageArray[1] = image;
					planarImage = JAI.create("AWTImage", image);
					localHistogramList = localHistogram.createLocalHistogram(planarImage);
					
					for(int m = 0; m < localHistogramList.length; m++){
						localHistogramArray[1][m] = (Histogram)localHistogramList[m];						
					}
				}
				
				// Picks frames between the first frame and the last frame
				else{
					framePick = (i*(main.mediaPlayer.getFrequency()));
					
					// Creates histogram 
					image = main.frameSelector.grabFrame(framePick);
					imageArray[1] = image;
					planarImage = JAI.create("AWTImage", image);
					localHistogramList = localHistogram.createLocalHistogram(planarImage);
					
					for(int m = 0; m < localHistogramList.length; m++){						
						localHistogramArray[1][m] = (Histogram)localHistogramList[m];						
					}
				}
				
				// Checks if the histogram difference is greater than the threshold value for a cut				
				if(i > 0){
					int matchingRegions = 0;
					
					// Checks all the regions in the picture					
					for(int n = 0; n < 9; n++){
						
						if(main.localHistogramComparator.compareLocalHistograms((Histogram)localHistogramArray[0][n], (Histogram)localHistogramArray[1][n])){
							matchingRegions++;
							
							// Performs the image view operations during the last iteration
							if(n == 8){								
								// More than 7 regions have to match before a cut is declared
								if(matchingRegions > 7){
									BufferedImage bufferedImage1 = main.imageResizer.resizeImage((Image)imageArray[0], imageWidth, imageHeight);
									BufferedImage bufferedImage2 = main.imageResizer.resizeImage((Image)imageArray[1], imageWidth, imageHeight);
									resizedImageObject1 = new ResizedImageObject(bufferedImage1, i-1, (i-1)*(main.mediaPlayer.getFrequency()));
									resizedImageObject2 = new ResizedImageObject(bufferedImage2, i, (i)*(main.mediaPlayer.getFrequency()));
															
									// Adds the remaining cut images
									main.gui.addImage(resizedImageObject2);
									imageCounter++;
								}
							}
						}
						
						// Sets the shot window properties after the last iteration
						if(i == ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond()))){
													
							// Sets the number of rows in the GUI
							double doubleImageCounter = (double)imageCounter;						
							double numberOfRows = (doubleImageCounter+1)/7;						
							double rowsCeil = Math.ceil(numberOfRows);						
							int integerNumberOfRows = (int)rowsCeil;						
							
							int imagePanelWidth;
							
							if(imageCounter >= 6){
								imagePanelWidth = (((resizedImageObject1.getBufferedImage().getWidth())*7)+38);							
							}
							
							else{
								imagePanelWidth = (((resizedImageObject1.getBufferedImage().getWidth())*(imageCounter+1))+38);							
							}
							
							int imagePanelHeight = imagePanelHeight = (((resizedImageObject1.getBufferedImage().getHeight())*integerNumberOfRows));
							
							main.gui.gridLayout.setRows(integerNumberOfRows);
							main.gui.imagePanel.setSize(imagePanelWidth, imagePanelHeight);
							
							// Sets the scrolling in the GUI if necessary
							if(main.gui.imagePanel.getHeight() >= 575){
								main.shotWindow.setSize(main.gui.imagePanel.getWidth(), 575);
							}
							else{
								main.shotWindow.setSize(main.gui.imagePanel.getWidth(), main.gui.imagePanel.getHeight()+30);							
							}
							
							// Makes the shot window visible						
							main.gui.shotWindow.setTitle("Video shot viewer - This video contains "+(imageCounter+1)+" shots");
							main.shotWindow.setVisible(true);
						}						
					}
					
					// Changes objects for the next iteration
					imageArray[0] = (Image)imageArray[1];					
					for(int o = 0; o < 9; o++){						
						localHistogramArray[0][o] = (Histogram)localHistogramArray[1][o];
					}
				}																
			}									
		}
		
		// The global edge chasing algorithm
		else if(main.algorithm3){
			
			for(int i = 0; i <= ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond())); i++){
				
				// Picks out the first frame, the last frame and a given number of frames per second									
				// Picks the first frame
				if(i == 0){
					framePick = 1;
					image = main.frameSelector.grabFrame(framePick);
					imageArray[0] = image;
					planarImage = JAI.create("AWTImage", image);
					edgeImage = edgeDetector.detectEdges(planarImage);
					edgeHistogramList = edgeHistogram.createEdgeHistogram(edgeImage);
					
					for(int m = 0; m < edgeHistogramList.length; m++){
						edgeHistogramArray[0][m] = (Histogram)edgeHistogramList[m];						
					}
					
					// Sets the image stats
					imageWidth = planarImage.getWidth();
					imageHeight = planarImage.getHeight();
					numberOfPixels = imageWidth*imageHeight;
					edgeColorThreshold = numberOfPixels/2200;
					
					// Adds the first image to the GUI
					BufferedImage bufferedImage1 = main.imageResizer.resizeImage((Image)imageArray[0], imageWidth, imageHeight);
					resizedImageObject1 = new ResizedImageObject(bufferedImage1, 1, 1);
					main.gui.addImage(resizedImageObject1);					
				}
				
				// Picks the last frame
				else if(i == ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond()))){
					framePick = (i*(main.mediaPlayer.getFrequency()));
										
					// Creates histogram 
					image = main.frameSelector.grabFrame(framePick);
					imageArray[1] = image;
					planarImage = JAI.create("AWTImage", image);
					edgeImage = edgeDetector.detectEdges(planarImage);					
					edgeHistogramList = edgeHistogram.createEdgeHistogram(edgeImage);
					
					for(int m = 0; m < edgeHistogramList.length; m++){
						edgeHistogramArray[1][m] = (Histogram)edgeHistogramList[m];						
					}
				}
				
				// Picks frames between the first frame and the last frame
				else{
					framePick = (i*(main.mediaPlayer.getFrequency()));
					
					// Creates histogram 
					image = main.frameSelector.grabFrame(framePick);
					imageArray[1] = image;
					planarImage = JAI.create("AWTImage", image);					
					edgeImage = edgeDetector.detectEdges(planarImage);					
					edgeHistogramList = edgeHistogram.createEdgeHistogram(edgeImage);
					
					for(int m = 0; m < edgeHistogramList.length; m++){						
						edgeHistogramArray[1][m] = (Histogram)edgeHistogramList[m];						
					}
				}
				
				// Checks if the histogram difference is greater than the threshold value for a cut				
				if(i > 0){
					int matchingRegions = 0;
					
					// Checks all the regions in the picture					
					for(int n = 0; n < 9; n++){
						
						if(main.edgeHistogramComparator.compareEdgeHistograms((Histogram)edgeHistogramArray[0][n], (Histogram)edgeHistogramArray[1][n])){
							matchingRegions++;
							
							// Performs the image view operations during the last iteration
							if(n == 8){								
								// More than 7 regions have to match before a cut is declared
								if(matchingRegions > 7){
									BufferedImage bufferedImage1 = main.imageResizer.resizeImage((Image)imageArray[0], imageWidth, imageHeight);
									BufferedImage bufferedImage2 = main.imageResizer.resizeImage((Image)imageArray[1], imageWidth, imageHeight);									
									resizedImageObject1 = new ResizedImageObject(bufferedImage1, i-1, (i-1)*(main.mediaPlayer.getFrequency()));
									resizedImageObject2 = new ResizedImageObject(bufferedImage2, i, (i)*(main.mediaPlayer.getFrequency()));
																		
									// Adds the remaining cut images
									main.gui.addImage(resizedImageObject2);									
									imageCounter++;								
								}
							}
						}
						
						// Sets the shot window properties after the last iteration
						if(i == ((main.mediaPlayer.getDurationInSeconds())*(main.mediaPlayer.getFramesPerSecond()))){
													
							// Sets the number of rows in the GUI
							double doubleImageCounter = (double)imageCounter;						
							double numberOfRows = (doubleImageCounter+1)/7;						
							double rowsCeil = Math.ceil(numberOfRows);						
							int integerNumberOfRows = (int)rowsCeil;						
							
							int imagePanelWidth;
							
							if(imageCounter >= 6){
								imagePanelWidth = (((resizedImageObject1.getBufferedImage().getWidth())*7)+38);							
							}
							
							else{
								imagePanelWidth = (((resizedImageObject1.getBufferedImage().getWidth())*(imageCounter+1))+38);							
							}
							
							int imagePanelHeight = imagePanelHeight = (((resizedImageObject1.getBufferedImage().getHeight())*integerNumberOfRows));
							
							main.gui.gridLayout.setRows(integerNumberOfRows);
							main.gui.imagePanel.setSize(imagePanelWidth, imagePanelHeight);
							
							// Sets the scrolling in the GUI if necessary
							if(main.gui.imagePanel.getHeight() >= 575){
								main.shotWindow.setSize(main.gui.imagePanel.getWidth(), 575);
							}
							else{
								main.shotWindow.setSize(main.gui.imagePanel.getWidth(), main.gui.imagePanel.getHeight()+30);							
							}
							
							// Makes the shot window visible						
							main.gui.shotWindow.setTitle("Video shot viewer - This video contains "+(imageCounter+1)+" shots");
							main.shotWindow.setVisible(true);
						}						
					}
					
					// Changes objects for the next iteration
					imageArray[0] = (Image)imageArray[1];					
					for(int o = 0; o < 9; o++){						
						edgeHistogramArray[0][o] = (Histogram)edgeHistogramArray[1][o];
					}					
				}																
			}			
		}		
	}
		
	/** Method that returns the global color histogram threshold
	 * 
	 *  @return globalColorThreshold The global color histogram algorithm threshold
	 */
	public double getGlobalThreshold(){
		return globalColorThreshold;
	}
		
	/** Method that returns the local color histogram threshold
	 *  
	 *  @return localColorThreshold The local color histogram algorithm threshold
	 */
	public double getLocalThreshold(){
		return localColorThreshold;
	}
	
	/** Method that returns the local color histogram threshold
	 * 
	 *  @return edgeColorThreshold The edge chasing algorithm color threshold
	 */
	public double getEdgeThreshold(){
		return edgeColorThreshold;
	}
}
