import java.awt.BorderLayout;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.awt.*;
import javax.media.ControllerEvent;
import javax.media.ControllerListener;
import javax.media.Manager;
import javax.media.NoPlayerException;
import javax.media.*;
import javax.media.Player;
import javax.media.PrefetchCompleteEvent;
import javax.media.RealizeCompleteEvent;
import javax.media.control.FrameGrabbingControl;
import javax.media.control.FramePositioningControl;
import javax.swing.JInternalFrame;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;

/** Class that handles the media player to load video files
 *  
 *  @version 1.0 01.06.06
 *  @author Vegard Kosmo
 *  @see Videosegmenter
 *  @see GUI
 *  
 */
public class Mediaplayer extends JInternalFrame implements ControllerListener{
	private static final long serialVersionUID = 1;
	
	// Main application reference
	Videosegmenter main;
	
	// GUI reference
	GUI gui;
	
	// Media player variables
	public Player mediaPlayer;
	private URL url;
	private Dimension playerSize;
	private Point playerLocation;
	private String playerTitle;
	public FramePositioningControl fpc;
    public FrameGrabbingControl fgc;
    private Time duration;
    private int durationInSeconds;
    private int totalNumberOfFrames;
    private int framesPerSecond;
    private int frequency;
    
	/** Standard constructor 
	 *  
	 *  @param main Main class reference
	 *  @param gui GUI reference
	 */
	public Mediaplayer(Videosegmenter main, GUI gui){
		this.main = main;
		this.gui = gui;
		playerSize = new Dimension(360, 360);
		playerLocation = new Point(0, 40);
		playerTitle = "Video input";
		framesPerSecond = 1;
	}
	
	/** Method that builds a media player from an input video file
	 * 
	 *  @param fileName The file name of the video file
	 */
	public void buildPlayer(String fileName){
				
		// Creates an URL from the file name
		try{
			url = new URL(fileName);
						
			// Creates a video player
		    try {
		    	mediaPlayer = Manager.createPlayer(url);
		    } 
		    
		    catch (NoPlayerException e) {
				System.out.println("No player exception: " + e);
		    }
		}
		
		catch(MalformedURLException e){
			System.out.println("Malformed URL Exception: ");
			e.printStackTrace();
		}
		
		catch(IOException e){
			System.out.println("IO Exception: ");
			e.printStackTrace();
		}
		
		// Sets the player properties
		getContentPane().setLayout( new BorderLayout() );
		setSize(playerSize.width, playerSize.height);
		setLocation(playerLocation);
		setTitle(playerTitle);
		setVisible(true);
		mediaPlayer.addControllerListener((ControllerListener)this);
		mediaPlayer.realize();
				
		// Internal frame listener
		addInternalFrameListener(new InternalFrameAdapter() {
		    public void internalFrameClosing(InternalFrameEvent ife) {
		    	mediaPlayer.close();
		    }
		} );
				
		gui.desktopPane.add(this);
	}
	
	// Methods that handle events from the media player
	public void controllerUpdate(ControllerEvent ce) {
		
		// Setting the states of the media player
		if(ce instanceof RealizeCompleteEvent) {
			// Processes as much data as necessary to reduce the Controller's start latency to the shortest possible time
			if(mediaPlayer != null){
				mediaPlayer.prefetch();
			}
		}
		
		else if(ce instanceof PrefetchCompleteEvent){
		    int playerWidth;
		    int playerHeight;
			
		    // Frame grabbing 
		    fpc = (FramePositioningControl) mediaPlayer.getControl("javax.media.control.FramePositioningControl");
			fgc = (FrameGrabbingControl) mediaPlayer.getControl("javax.media.control.FrameGrabbingControl");
		    
		    if ((mediaPlayer.getVisualComponent()) != null) {
		    	playerWidth = 393;
		    	playerHeight = 355;
		    	getContentPane().add("Center", mediaPlayer.getVisualComponent());
		    } 
		    
		    // Sets the size if visual component is null
		    else{
		    	playerWidth = 360;
		    	playerHeight = 355;		    	
		    }
		    
		    if ((mediaPlayer.getControlPanelComponent()) != null) {
		    	getContentPane().add("South", mediaPlayer.getControlPanelComponent());
		    }
		    
		    // Sets the media player size
		    setSize(playerWidth, playerHeight);
		    validate();
		    
		    // The video file properties
			duration = mediaPlayer.getDuration();
			durationInSeconds = (int)duration.getSeconds();
			totalNumberOfFrames = fpc.mapTimeToFrame(duration);
			frequency = ((totalNumberOfFrames/durationInSeconds)/framesPerSecond);
			
			// Starts the given segmentation algorithm
			main.buildFrameSelector();
			main.algorithm = new Algorithm(main);
			main.algorithm.runAlgorithm();
		}
		
		// Handles the event from the media player start button
		else if(ce instanceof StartEvent){
			
		}
		
		else if(ce instanceof StopEvent){
			// Sets the player time to zero and stops the running algorithm
			mediaPlayer.setMediaTime(new Time(0.0)); 
		}
		

		else if(ce instanceof EndOfMediaEvent) {
			mediaPlayer.setMediaTime(new Time(0.0));
		}
	}

	/** Method to return the video file duration
	 * 
	 *  @return durationInSeconds The video file duration
	 */
	public int getDurationInSeconds(){
		return durationInSeconds;
	}
	
	/** Method to return the total number of frames in the video file
	 *  
	 *  @return totalNumberOfFrames The total number of frames in the video file
	 */
	public int getTotalNumberOfFrames(){
		return totalNumberOfFrames;
	}
	
	// 
	
	/** Method to return the frames per second value in the video file
	 *  
	 *  @return framesPerSecond The number of frames per second in the video file
	 */
	public int getFramesPerSecond(){
		return framesPerSecond;
	}
	
	/** Method to return the video file frequency
	 * 
	 *  @return frequency Returns the video file frequency
	 */
	public int getFrequency(){
		return frequency;
	}
	
	/** Method to start the media player on a given starting point
	 * 
	 *  @param startingPoint The point in the video file to start the video player
	 */
	public void startMediaPlayer(int startingPoint){
		fpc.seek(startingPoint);
		mediaPlayer.start();
	}
	
}
